#ifndef __ASCII_API_H__
#define __ASCII_API_H__

/* rc prefix stands for Roaming Components */

#ifdef __cplusplus
extern "C" {
#endif

#define ERR_GENERIC				-100

/* ASN.1 INTEGERs are stored as 64-bit integers */
#ifdef _WIN32
typedef long long RCINT64;
#elif __GNUC__
typedef long long RCINT64;
#endif

/* Handle to any ASN.1 node in the TAP tree */
typedef void* RCHANDLE;

/* Handle to XML template file */
typedef void* XMLHANDLE;

/* Handle to Visitor engine */
typedef void* RC_VISIT_HANDLE;

/* Callbacks used from the DLL/.so */
typedef void (*PFNMODULECALLBACK)(RCHANDLE, RCHANDLE, RCINT64);
typedef struct tagCallback {
    const char *_pszTypename;
    PFNMODULECALLBACK _pCallback;
} Callback;

/* Tree access */

/* Ask a SEQUENCE_OF how many entities it contains */
int GetCount( RCHANDLE rcHandle, char *path, RCINT64* retVal);

/* Get the value of an INTEGER */
int GetInteger( RCHANDLE rcHandle, char *path, RCINT64* retVal);

/* Set the value of an INTEGER */
int SetInteger( RCHANDLE rcHandle, char *path, RCINT64 val);

/* Get the value of an OCTET_STRING (BCDStrings are auto-converted) */
int GetString( RCHANDLE rcHandle, char *path, char* destBuffer, int sizeOfDestBuffer, int *pNumberOfWrittenBytes);

/* Set the value of an OCTET_STRING (BCDStrings are auto-converted) */
int SetString( RCHANDLE rcHandle, char *path, char* srcBuffer, int numberOfBytesToWrite);

/* Create a child node under a SEQUENCE or a CHOICE */
int CreateChildVariable(RCHANDLE parent, char *variableName, RCHANDLE *child);

/* Follow a path to a specific node */
int NavigateAndObtainHandle( RCHANDLE parent, const char *path, RCHANDLE *destHandle);

/* Descend into an offspring node */
int NavigateToAChildVariable(RCHANDLE parent, char *variableName, RCHANDLE *child);

/* Append a new entry in a SEQUENCE_OF */
int AppendEntryInSequenceOf(RCHANDLE sequenceOf, RCHANDLE *child);

/* Navigate to a specific child of a SEQUENCE_OF (index based) */
int NavigateToAMemberOfASequenceOf(RCHANDLE parent, int idx, RCHANDLE *child);

/* Read a template XML and store it for re-use */
int CompileSubTreeFromXmlTemplate(char* xmlFileName, XMLHANDLE *subtree);

/* Apply a stored XML template to a specific node, generating an offspring */
int AddSubTree(RCHANDLE parent, XMLHANDLE subtree, RCHANDLE *offspring);

/* Look-up codes */

/* These functions search the appropriate places in the tree to find the element you
 * want; if not found, they add it and store the new code; otherwise they store the
 * existing code. The argument list to each function contains mandatory arguments
 * (starting with mand_) and optional ones (the ones representing CHOICEs) */

/* get_taxCode()
 * Input : taxType, taxRate (for TAP3.11, chargeType and taxIndicator as well - in 
 *         other versions pass NULL for chargeType)
 * Output: taxCode (found or just created)
 */
int get_taxCode(RCHANDLE di, const char * mand_taxType, const char * mand_taxRate, const char * mand_chargeType, const char * mand_taxIndicator, RCHANDLE *ptaxCode);

/* get_discountCode()
 * Input : fixedDiscountValue OR discountRate
 * Output: discountCode (found or just created)
 */
int get_discountCode(RCHANDLE di, RCINT64 *opt_DiscountApplied_fixedDiscountValue, RCINT64 *opt_DiscountApplied_discountRate, RCHANDLE *pdiscountCode);

/* get_exchangeRateCode()
 * Input  : numberOfDecimalPlaces, exchangeRate
 * Output : exchangeRateCode (found or just created)
 */
int get_exchangeRateCode(RCHANDLE di, RCINT64 mand_numberOfDecimalPlaces, RCINT64 mand_exchangeRate, RCHANDLE *pexchangeRateCode);

/* get_utcTimeOffsetCode()
 * Input  : utcTimeOffset
 * Output : utcTimeOffsetCode (found or just created)
 */
int get_utcTimeOffsetCode(RCHANDLE di, const char * mand_utcTimeOffset, RCHANDLE *putcTimeOffsetCode);

/* get_recEntityCode()
 * Input  : recEntityType (for TAP3.11, recEntityId as well)
 *          and where applicable, one of the following: mscId, msisdn, IPTextV4Address, IPTextV6Address, gmlcAddress
 */
int get_recEntityCode(
    RCHANDLE di, 
    RCINT64 mand_recEntityType, 
    const char *opt_RecEntityId_mscId, 
    const char *opt_RecEntityId_msisdn, 
    const char *opt_RecEntityId_GsnAddress_iPTextV4Address, 
    const char *opt_RecEntityId_GsnAddress_iPTextV6Address, 
    const char *opt_RecEntityId_gmlcAddress, 
    const char *mand_recEntityId,
    RCHANDLE *precEntityCode);

/* get_calledNumAnalysisCode()
 * Input  : countryCode, iac
 * Output : calledNumAnalysisCode (found or just created)
 */
int get_calledNumAnalysisCode(RCHANDLE di, const char * mand_CountryCode, const char * mand_Iac, RCHANDLE *pcalledNumAnalysisCode);

/* get_vasCode()
 * Input  : vasShortDesc, vasDesc
 * Output : vasCode (found or just created)
 */
int get_vasCode(RCHANDLE di, const char * mand_vasShortDesc, const char * mand_vasDesc, RCHANDLE *pvasCode);

/* get_messageDescriptionCode()
 * Input  : messageDescription
 * Output : messageDescriptionCode (found or just created)
 */
int get_messageDescriptionCode(RCHANDLE di, const char * mand_messageDescription, RCHANDLE *pmessageDescriptionCode);

/* The only valid way to report an error: pass a null-terminated string describing what went wrong */
void throwException(const char *);

/* Adaptive programming inspired visitors */

/* Setup a recursive traversal of the tree from the parent, searching for all nodes of Type 'type' */
RC_VISIT_HANDLE InitializeVisit( RCHANDLE parent, char *type );

/* Execute the traversal until you find the next node, or until you exhaust them - return NULL if no
 * other node of type 'type' can be found (see InitializeVisit) */
RCHANDLE GetNext( RC_VISIT_HANDLE );

/* Abort a recursive traversal midway (before NULL is returned from GetNext() */
void DestroySearch(RC_VISIT_HANDLE);

/* Structure used to migrate the callback function pointers over to the module */
typedef struct tagUtils {

    /* for help with these methods, read the comments above */
    int (*pfnGetCount)   ( RCHANDLE rcHandle, char *path, RCINT64* retVal);
    int (*pfnGetInteger) ( RCHANDLE rcHandle, char *path, RCINT64* retVal);
    int (*pfnSetInteger) ( RCHANDLE rcHandle, char *path, RCINT64 val);
    int (*pfnGetString)  ( RCHANDLE rcHandle, char *path, char* destBuffer, int sizeOfDestBuffer, int *pNumberOfWrittenBytes);
    int (*pfnSetString)  ( RCHANDLE rcHandle, char *path, char* srcBuffer, int numberOfBytesToWrite);
    int (*pfnNavigateAndObtainHandle) ( RCHANDLE parent, const char *path, RCHANDLE *destHandle);
    int (*pfnCreateChildVariable) (RCHANDLE parent, char *variableName, RCHANDLE *child);
    int (*pfnNavigateToAChildVariable) (RCHANDLE parent, char *variableName, RCHANDLE *child);
    int (*pfnAppendEntryInSequenceOf) (RCHANDLE sequenceOf, RCHANDLE *child);
    int (*pfnNavigateToAMemberOfASequenceOf) (RCHANDLE parent, int idx, RCHANDLE *child);
    int (*pfnCompileSubTreeFromXmlTemplate) (char* xmlFileName, XMLHANDLE *subtree);
    int (*pfnAddSubTree) (RCHANDLE parent, XMLHANDLE subtree, RCHANDLE *offspring);
    RC_VISIT_HANDLE (*pfnInitializeVisit) ( RCHANDLE parent, char *type );
    RCHANDLE        (*pfnGetNext)         ( RC_VISIT_HANDLE );
    void (*pfnDestroySearch) (RC_VISIT_HANDLE);

    int (*pfn_get_taxCode) (RCHANDLE di, const char * mand_taxType, const char * mand_taxRate, const char * mand_chargeType, const char * mand_taxIndicator, RCHANDLE *ptaxCode);
    int (*pfn_get_discountCode) (RCHANDLE di, RCINT64 *opt_DiscountApplied_fixedDiscountValue, RCINT64 *opt_DiscountApplied_discountRate, RCHANDLE *pdiscountCode);
    int (*pfn_get_exchangeRateCode) (RCHANDLE di, RCINT64 mand_numberOfDecimalPlaces, RCINT64 mand_exchangeRate, RCHANDLE *pexchangeRateCode);
    int (*pfn_get_utcTimeOffsetCode) (RCHANDLE di, const char * mand_utcTimeOffset, RCHANDLE *putcTimeOffsetCode);
    int (*pfn_get_recEntityCode)(
	RCHANDLE di, 
	RCINT64 mand_recEntityType, 
	const char *opt_RecEntityId_mscId, 
	const char *opt_RecEntityId_msisdn, 
	const char *opt_RecEntityId_GsnAddress_iPTextV4Address, 
	const char *opt_RecEntityId_GsnAddress_iPTextV6Address, 
	const char *opt_RecEntityId_gmlcAddress, 
	const char *mand_recEntityId,
	RCHANDLE *precEntityCode);
    int (*pfn_get_calledNumAnalysisCode) (RCHANDLE di, const char * mand_CountryCode, const char * mand_Iac, RCHANDLE *pcalledNumAnalysisCode);
    int (*pfn_get_vasCode) (RCHANDLE di, const char * mand_vasShortDesc, const char * mand_vasDesc, RCHANDLE *pvasCode);
    int (*pfn_get_messageDescriptionCode) (RCHANDLE di, const char * mand_messageDescription, RCHANDLE *pmessageDescriptionCode);

    void (*pfnThrowException) (const char *);
} Utils;

typedef Callback* (*PFNGetCB)(Utils*, char *, char *);

#ifdef __cplusplus
}
#endif

#endif
